//   The contents of this file are subject to the Mozilla Public License
//   Version 1.1 (the "License"); you may not use this file except in
//   compliance with the License. You may obtain a copy of the License at
//   http://www.mozilla.org/MPL/
//
//   Software distributed under the License is distributed on an "AS IS"
//   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//   License for the specific language governing rights and limitations
//   under the License.
//
//   The Original Code is RabbitMQ.
//
//   The Initial Developers of the Original Code are LShift Ltd.,
//   Cohesive Financial Technologies LLC., and Rabbit Technologies Ltd.
//
//   Portions created by LShift Ltd., Cohesive Financial Technologies
//   LLC., and Rabbit Technologies Ltd. are Copyright (C) 2007-2008
//   LShift Ltd., Cohesive Financial Technologies LLC., and Rabbit
//   Technologies Ltd.;
//
//   All Rights Reserved.
//
//   Contributor(s): ______________________________________.
//

package com.rabbitmq.client.impl;

import java.io.DataOutputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Date;
import java.util.Map;

import org.apache.commons.io.IOUtils;


/**
 * Generates AMQP wire-protocol encoded arguments. Methods on this
 * object are usually called from autogenerated code.
 */
public class MethodArgumentWriter
{
    /** Accumulates our output */
    private final DataOutputStream out;
    /** When encoding one or more bits, records whether a group of bits is waiting to be written */
    private boolean needBitFlush;
    /** The current group of bits */
    private byte bitAccumulator;
    /** The current position within the group of bits */
    private int bitMask;

    /**
     * Constructs a MethodArgumentWriter targetting the given DataOutputStream.
     */
    public MethodArgumentWriter(DataOutputStream out)
    {
        this.out = out;
        resetBitAccumulator();
    }

    /** Private API - called to reset the bit group variables. */
    private void resetBitAccumulator() {
        needBitFlush = false;
        bitAccumulator = 0;
        bitMask = 1;
    }
    
    /**
     * Private API - called when we may be transitioning from encoding
     * a group of bits to encoding a non-bit value.
     */
    private final void bitflush() 
        throws IOException
    {
        if (needBitFlush) {
            out.writeByte(bitAccumulator);
            resetBitAccumulator();
        }
    }
    
    /** Public API - encodes a short string argument. */
    public final void writeShortstr(String str)
        throws IOException
    {
        bitflush();
        byte [] bytes = str.getBytes("utf-8");
        out.writeByte(bytes.length);
        out.write(bytes);
    }

    /** Public API - encodes a long string argument from a LongString. */
    public final void writeLongstr(LongString str)
        throws IOException
    {
        bitflush();
        writeLong((int)str.length());
        IOUtils.copy(str.getStream(), out);
    }

    /** Public API - encodes a long string argument from a String. */
    public final void writeLongstr(String str)
        throws IOException
    {
        bitflush();
        byte [] bytes = str.getBytes("utf-8");
        writeLong(bytes.length);
        out.write(bytes);
    }
    
    /** Public API - encodes a short integer argument. */
    public final void writeShort(int s)
        throws IOException
    {
        bitflush();
        out.writeShort(s);
    }

    /** Public API - encodes an integer argument. */
    public final void writeLong(int l)
        throws IOException
    {
        bitflush();
        // java's arithmetic on this type is signed, however its
        // reasonable to use ints to represent the unsigned long
        // type - for values < Integer.MAX_VALUE everything works
        // as expected
        out.writeInt(l);
    }

    /** Public API - encodes a long integer argument. */
    public final void writeLonglong(long ll) 
        throws IOException
    {
        bitflush();
        out.writeLong(ll);
    }

    /** Public API - encodes a boolean/bit argument. */
    public void writeBit(boolean b) 
        throws IOException
    {
        if (bitMask > 0x80) {
            bitflush();
        }
        if (b) {
            bitAccumulator |= bitMask;
        } else {
            // um, don't set the bit.
        }
        bitMask = bitMask << 1;
        needBitFlush = true;
    }

    /** Public API - encodes a table argument. */
    public final void writeTable(Map<String, Object> table)
        throws IOException
    {
        bitflush();
        if (table == null) {
            // Convenience.
            out.writeInt(0);
        } else {
            out.writeInt((int)Frame.tableSize(table));
            for(Map.Entry<String,Object> entry:  table.entrySet()) {
                writeShortstr(entry.getKey());
                Object value = entry.getValue();
                if(value instanceof String) {
                    writeOctet('S');
                    writeLongstr((String)value);
                }
                else if(value instanceof LongString) {
                    writeOctet('S');
                    writeLongstr((LongString)value);
                }
                else if(value instanceof Integer) {
                    writeOctet('I');
                    writeLong((Integer) value);
                }
                else if(value instanceof BigDecimal) {
                    writeOctet('D');
                    BigDecimal decimal = (BigDecimal)value;
                    writeOctet(decimal.scale());
                    BigInteger unscaled = decimal.unscaledValue();
                    if(unscaled.bitLength() > 32) /*Integer.SIZE in Java 1.5*/
                        throw new IllegalArgumentException
                            ("BigDecimal too large to be encoded");
                    writeLong(decimal.unscaledValue().intValue());
                }
                else if(value instanceof Date) {
                    writeOctet('T');
                    writeTimestamp((Date)value);
                }
                else if(value instanceof Map) {
                    writeOctet('F');
                    // Ignore the warnings here.  We hate erasure
                    // (not even a little respect)
                    writeTable((Map<String, Object>) value);
                }
                else {
                    throw new IllegalArgumentException
                        ("Invalid value type: " + value.getClass().getName()
                         + " for key " + entry.getKey());
                }
            }
        }
    }

    /** Public API - encodes an octet argument from an int. */
    public final void writeOctet(int octet) 
        throws IOException
    {
        bitflush();
        out.writeByte(octet);
    }
    
    /** Public API - encodes an octet argument from a byte. */
    public final void writeOctet(byte octet)
        throws IOException
    {
        bitflush();
        out.writeByte(octet);
    }

    /** Public API - encodes a timestamp argument. */
    public final void writeTimestamp(Date timestamp) 
        throws IOException
    {
        // AMQP uses POSIX time_t which is in seconds since the epoch began
        writeLonglong(timestamp.getTime()/1000);
    }

    /**
     * Public API - call this to ensure all accumulated argument
     * values are correctly written to the output stream.
     */
    public void flush() 
        throws IOException
    {
        bitflush();
        out.flush();
    }
}
